use std::ops::{Add, Sub, Mul, Div, Rem, Neg};
use std::ops::{AddAssign, SubAssign, MulAssign, DivAssign, RemAssign};

use ::constants::Constants;
use ::real::Real;



/// A degree usually denoted by ° (the degree symbol),
/// is a measurement of a plane angle, defined so that a
/// full rotation is 360 degrees.
#[derive(Clone, Copy, Eq, PartialEq, Hash)]
pub struct Degree<T>
{
   value: T
}

/// A unit of angle, equal to an angle at the center of
/// a circle whose arc is equal in length to the radius.
#[derive(Clone, Copy, Eq, PartialEq, Hash)]
pub struct Radian<T>
{
   value: T
}



impl<T> Degree<T> where T: Real
{
   /// Create a new Degree structure with the given value.
   pub fn new(val: T) -> Degree<T>
   {
      Degree
      {
         value: val
      }
   }

   /// Computes the arccosine of a number. Return value is in Degrees in
   /// the range [0, 180] or NaN if the number is outside the range
   /// [-1, 1].
   ///
   ///```
   /// use sigils::{Degree, Real};
   ///
   /// let f: Degree<f64>;
   ///
   /// f = Degree::from(45.0f64);
   /// assert!((45.0f64 - *Degree::acos(f.cos())) < 1e-10);
   ///```
   pub fn acos(val: T) -> Degree<T>
   {
      let radians: Radian<T>;

      radians = val.acos();
      Degree::from(radians)
   }

   /// Computes the arcsine of a number. Return value is in Degrees in
   /// the range [-57.2957795/2, 57.2957795/2] or NaN if the number is
   /// outside the range [-1, 1].
   ///
   ///```
   /// use sigils::{Degree, Real};
   ///
   /// let f: Degree<f64>;
   ///
   /// f = Degree::from(45.0f64);
   /// assert!((45.0f64 - *Degree::asin(f.sin())) < 1e-10);
   ///```
   pub fn asin(val: T) -> Degree<T>
   {
      let radians: Radian<T>;

      radians = val.asin();
      Degree::from(radians)
   }

   /// Computes the arctangent of a number. Return value is in degrees in the
   /// range [-57.2957795/2, 57.2957795/2];
   ///
   ///```
   /// use sigils::{Degree, Real};
   ///
   /// let f: Degree<f64>;
   ///
   /// f = Degree::from(45.0f64);
   /// assert!((45.0f64 - *Degree::atan(f.tan())) < 1e-10);
   ///```
   pub fn atan(val: T) -> Degree<T>
   {
      let radians: Radian<T>;

      radians = val.atan();
      Degree::from(radians)
   }

   /// Inverse hyperbolic cosine function.
   pub fn acosh(val: T) -> Degree<T>
   {
      let radians: Radian<T>;

      radians = val.acosh();
      Degree::from(radians)
   }

   /// Inverse hyperbolic sine function.
   pub fn asinh(val: T) -> Degree<T>
   {
      let radians: Radian<T>;

      radians = val.asinh();
      Degree::from(radians)
   }

   /// Inverse hyperbolic tangent function.
   pub fn atanh(val: T) -> Degree<T>
   {
      let radians: Radian<T>;

      radians = val.atanh();
      Degree::from(radians)
   }

   /// Computes the four quadrant arctangent of `self` (`y`) and `other` (`x`).
   pub fn atan2(y: T, x: T) -> Degree<T>
   {
      let radians: Radian<T>;

      radians = y.atan2(x);
      Degree::from(radians)
   }

   /// Computes the cosine of this angle.
   ///
   /// ```
   /// use sigils::Degree;
   /// use std::f64;
   ///
   /// let x: Degree<f64> = Degree::from(360.0f64);
   ///
   /// let abs_difference = (x.cos() - 1.0).abs();
   ///
   /// assert!(abs_difference < 1e-10);
   /// ```
   pub fn cos(&self) -> T
   {
      let radians: Radian<T>;

      radians = Radian::from(*self);
      radians.cos()
   }

   /// Computes the sine of this angle.
   ///
   /// ```
   /// use sigils::Degree;
   /// use std::f64;
   ///
   /// let x: Degree<f64> = Degree::from(90.0f64);
   ///
   /// let abs_difference = (x.sin() - 1.0).abs();
   ///
   /// assert!(abs_difference < 1e-10);
   /// ```
   pub fn sin(&self) -> T
   {
      let radians: Radian<T>;

      radians = Radian::from(*self);
      radians.sin()
   }

   /// Computes the tangent of this angle.
   ///
   /// ```
   /// use sigils::Degree;
   /// use std::f64;
   ///
   /// let x: Degree<f64> = Degree::from(45.0f64);
   ///
   /// let abs_difference = (x.tan() - 1.0).abs();
   ///
   /// assert!(abs_difference < 1e-14);
   /// ```
   pub fn tan(&self) -> T
   {
      let radians: Radian<T>;

      radians = Radian::from(*self);
      radians.tan()
   }

   /// Hyperbolic cosine function.
   ///
   /// ```
   /// use sigils::Degree;
   /// use sigils::Constants;
   ///
   /// let e32: f32 = Constants::E;
   /// let x32: Degree<f32> = Degree::from(f32::INVERSE_PI_DIVIDED_BY_180);
   /// let f_val32 = x32.cosh();
   /// let g_val32 = (e32*e32 + 1.0f32)/(2.0f32*e32);
   /// let abs_difference32 = (f_val32 - g_val32).abs();
   ///
   /// let e64: f64 = Constants::E;
   /// let x64: Degree<f64> = Degree::from(f64::INVERSE_PI_DIVIDED_BY_180);
   /// let f_val64 = x64.cosh();
   /// let g_val64 = (e64*e64 + 1.0f64)/(2.0f64*e64);
   /// let abs_difference64 = (f_val64 - g_val64).abs();
   ///
   /// // Solving cosh() at 1 gives this result
   /// //assert!(abs_difference32 < 1.0e-10);
   /// assert!(abs_difference64 < 1.0e-10);
   /// ```
   pub fn cosh(&self) -> T
   {
      let radians: Radian<T>;

      radians = Radian::from(*self);
      radians.cosh()
   }

   /// Hyperbolic sine function.
   ///
   /// ```
   /// use sigils::Degree;
   /// use sigils::Constants;
   ///
   /// let e32: f32 = Constants::E;
   /// let x32: Degree<f32> = Degree::from(f32::INVERSE_PI_DIVIDED_BY_180);
   ///
   /// let f_val32 = x32.sinh();
   /// let g_val32 = (e32*e32 - 1.0f32)/(2.0f32*e32);
   /// let abs_difference32 = (f_val32 - g_val32).abs();
   ///
   /// let e64: f64 = Constants::E;
   /// let x64: Degree<f64> = Degree::from(f64::INVERSE_PI_DIVIDED_BY_180);
   ///
   /// let f_val64 = x64.sinh();
   /// let g_val64 = (e64*e64 - 1.0f64)/(2.0f64*e64);
   /// let abs_difference64 = (f_val64 - g_val64).abs();
   ///
   /// // Solving sinh() at 1 gives `(e^2-1)/(2e)`
   /// //assert!(abs_difference32 < 1e-10);
   /// assert!(abs_difference64 < 1e-10);
   /// ```
   pub fn sinh(&self) -> T
   {
      let radians: Radian<T>;

      radians = Radian::from(*self);
      radians.sinh()
   }

   /// Hyperbolic tangent function.
   ///
   /// ```
   /// use sigils::Degree;
   /// use sigils::Constants;
   ///
   /// let e32: f32 = Constants::E;
   /// let x32: Degree<f32> = Degree::from(f32::INVERSE_PI_DIVIDED_BY_180);
   ///
   /// let f_val32 = x32.tanh();
   /// let g_val32 = (1.0f32 - e32.powi(-2i32))/(1.0f32 + e32.powi(-2i32));
   /// let abs_difference32 = (f_val32 - g_val32).abs();
   ///
   /// let e64: f64 = Constants::E;
   /// let x64: Degree<f64> = Degree::from(f64::INVERSE_PI_DIVIDED_BY_180);
   ///
   /// let f_val64 = x64.tanh();
   /// let g_val64 = (1.0f64 - e64.powi(-2i32))/(1.0f64 + e64.powi(-2i32));
   /// let abs_difference64 = (f_val64 - g_val64).abs();
   ///
   /// // Solving tanh() at 1 gives `(1 - e^(-2))/(1 + e^(-2))`
   /// //assert!(abs_difference32 < 1.0e-10);
   /// assert!(abs_difference64 < 1.0e-10);
   /// ```
   pub fn tanh(&self) -> T
   {
      let radians: Radian<T>;

      radians = Radian::from(*self);
      radians.tanh()
   }
}

impl<T> ::std::ops::Deref for Degree<T> where T: Real
{
   type Target = T;

   fn deref(&self) -> &T
   {
      &self.value
   }
}

impl<T> ::std::ops::DerefMut for Degree<T> where T: Real
{
   fn deref_mut<'a>(&'a mut self) -> &'a mut T
   {
      &mut self.value
   }
}

binary_operator_impl!(Add::add, Degree {value}, Real);
binary_operator_impl!(Sub::sub, Degree {value}, Real);
binary_operator_impl!(Mul::mul, Degree {value}, Real);
binary_operator_impl!(Div::div, Degree {value}, Real);
binary_operator_impl!(Rem::rem, Degree {value}, Real);

impl<T> ::std::ops::Neg for Degree<T>
   where T: Neg<Output = T>
{
   type Output = Degree<T>;

   fn neg(self) -> Degree<T>
   {
      let mut degrees: Degree<T>;

      degrees = self;
      degrees.value = degrees.value.neg();
      degrees
   }
}

binary_operator_assign_impl!(AddAssign::add_assign,
                             Degree {value}, Real);
binary_operator_assign_impl!(SubAssign::sub_assign,
                             Degree {value}, Real);
binary_operator_assign_impl!(MulAssign::mul_assign,
                             Degree {value}, Real);
binary_operator_assign_impl!(DivAssign::div_assign,
                             Degree {value}, Real);
binary_operator_assign_impl!(RemAssign::rem_assign,
                             Degree {value}, Real);

impl ::std::convert::From<Degree<f32>> for f32
{
   fn from(degrees: Degree<f32>) -> f32
   {
      degrees.value
   }
}

impl ::std::convert::From<Degree<f64>> for f64
{
   fn from(degrees: Degree<f64>) -> f64
   {
      degrees.value
   }
}

impl<T> ::std::convert::From<T> for Degree<T> where T: Real
{
   fn from(val: T) -> Degree<T>
   {
      Degree::new(val)
   }
}

impl<T> ::std::convert::From<Radian<T>> for Degree<T> where T: Real
{
   fn from(radians: Radian<T>) -> Degree<T>
   {
      let degs: T;

      degs = radians.value * Constants::INVERSE_PI_DIVIDED_BY_180;
      Degree::new(degs)
   }
}

impl<T> ::std::fmt::Debug for Degree<T>
   where T: Real
{
   fn fmt(&self, f: &mut ::std::fmt::Formatter) -> ::std::fmt::Result
   {
      ::std::fmt::Display::fmt(self, f)
   }
}

impl<T> ::std::fmt::Display for Degree<T>
   where T: Real
{
   fn fmt(&self, f: &mut ::std::fmt::Formatter) -> ::std::fmt::Result
   {
      write!(f, "{}", self.value)
   }
}

impl<T> Radian<T> where T: Real
{
   /// Create a new Radian structure with the given value.
   pub fn new(val: T) -> Radian<T>
   {
      Radian
      {
         value: val
      }
   }

   /// Computes the arccosine of a number. Return value is in Degrees in
   /// the range [0, pi] or NaN if the number is outside the range
   /// [-1, 1].
   ///
   ///```
   /// use sigils::{Constants, Radian, Real};
   ///
   /// let f: Radian<f64>;
   ///
   /// f = Radian::from(f64::PI / 4.0f64);
   /// assert!(((f64::PI / 4.0f64) - *Radian::acos(f.cos())) < 1e-10);
   ///```
   pub fn acos(val: T) -> Radian<T>
   {
      val.acos()
   }

   /// Computes the arcsine of a number. Return value is in Degrees in
   /// the range [-pi/2, pi/2] or NaN if the number is
   /// outside the range [-1, 1].
   ///
   ///```
   /// use sigils::{Constants, Radian, Real};
   ///
   /// let f: Radian<f64>;
   ///
   /// f = Radian::from(f64::PI / 4.0f64);
   /// assert!(((f64::PI / 4.0f64) - *Radian::asin(f.sin())) < 1e-10);
   ///```
   pub fn asin(val: T) -> Radian<T>
   {
      val.asin()
   }

   /// Computes the arctangent of a number. Return value is in degrees in the
   /// range [-pi/2, pi/2];
   ///
   ///```
   /// use sigils::{Constants, Radian, Real};
   ///
   /// let f: Radian<f64>;
   ///
   /// f = Radian::from(f64::PI / 4.0f64);
   /// assert!(((f64::PI / 4.0f64) - *Radian::atan(f.tan())) < 1e-10);
   ///```
   pub fn atan(val: T) -> Radian<T>
   {
      val.atan()
   }

   /// Inverse hyperbolic cosine function.
   pub fn acosh(val: T) -> Radian<T>
   {
      val.acosh()
   }

   /// Inverse hyperbolic sine function.
   pub fn asinh(val: T) -> Radian<T>
   {
      val.asinh()
   }

   /// Inverse hyperbolic tangent function.
   pub fn atanh(val: T) -> Radian<T>
   {
      val.atanh()
   }

   /// Computes the four quadrant arctangent of `self` (`y`) and `other` (`x`).
   pub fn atan2(y: T, x: T) -> Radian<T>
   {
      y.atan2(x)
   }

   /// Computes the cosine of this angle.
   ///
   /// ```
   /// use sigils::Radian;
   /// use std::f64;
   ///
   /// let x: Radian<f64> = Radian::from(2.0*f64::consts::PI);
   ///
   /// let abs_difference = (x.cos() - 1.0).abs();
   ///
   /// assert!(abs_difference < 1e-10);
   /// ```
   pub fn cos(&self) -> T
   {
      self.value.cos()
   }

   /// Computes the sine of this angle.
   ///
   /// ```
   /// use sigils::Radian;
   /// use std::f64;
   ///
   /// let x: Radian<f64> = Radian::from(f64::consts::PI/2.0);
   ///
   /// let abs_difference = (x.sin() - 1.0).abs();
   ///
   /// assert!(abs_difference < 1e-10);
   /// ```
   pub fn sin(&self) -> T
   {
      self.value.sin()
   }

   /// Computes the tangent of this angle.
   ///
   /// ```
   /// use sigils::Radian;
   /// use std::f64;
   ///
   /// let x: Radian<f64> = Radian::from(f64::consts::PI/4.0);
   ///
   /// let abs_difference = (x.tan() - 1.0).abs();
   ///
   /// assert!(abs_difference < 1e-14);
   /// ```
   pub fn tan(&self) -> T
   {
      self.value.tan()
   }

   /// Hyperbolic cosine function.
   ///
   /// ```
   /// use sigils::Radian;
   /// use sigils::Constants;
   ///
   /// let e32: f32 = Constants::E;
   /// let x32: Radian<f32> = Radian::new(1.0f32);
   /// let f_val32 = x32.cosh();
   /// let g_val32 = (e32*e32 + 1.0f32)/(2.0f32*e32);
   /// let abs_difference32 = (f_val32 - g_val32).abs();
   ///
   /// let e64: f64 = Constants::E;
   /// let x64: Radian<f64> = Radian::new(1.0f64);
   /// let f_val64 = x64.cosh();
   /// let g_val64 = (e64*e64 + 1.0f64)/(2.0f64*e64);
   /// let abs_difference64 = (f_val64 - g_val64).abs();
   ///
   /// // Solving cosh() at 1 gives this result
   /// //assert!(abs_difference32 < 1.0e-10);
   /// assert!(abs_difference64 < 1.0e-10);
   /// ```
   pub fn cosh(&self) -> T
   {
      self.value.cosh()
   }

   /// Hyperbolic sine function.
   ///
   /// ```
   /// use sigils::Radian;
   /// use sigils::Constants;
   ///
   /// let e32: f32 = Constants::E;
   /// let x32: Radian<f32> = Radian::from(1.0f32);
   ///
   /// let f_val32 = x32.sinh();
   /// let g_val32 = (e32*e32 - 1.0f32)/(2.0f32*e32);
   /// let abs_difference32 = (f_val32 - g_val32).abs();
   ///
   /// let e64: f64 = Constants::E;
   /// let x64: Radian<f64> = Radian::from(1.0f64);
   ///
   /// let f_val64 = x64.sinh();
   /// let g_val64 = (e64*e64 - 1.0f64)/(2.0f64*e64);
   /// let abs_difference64 = (f_val64 - g_val64).abs();
   ///
   /// // Solving sinh() at 1 gives `(e^2-1)/(2e)`
   /// //assert!(abs_difference32 < 1e-10);
   /// assert!(abs_difference64 < 1e-10);
   /// ```
   pub fn sinh(&self) -> T
   {
      self.value.sinh()
   }

   /// Hyperbolic tangent function.
   ///
   /// ```
   /// use sigils::Radian;
   /// use sigils::Constants;
   ///
   /// let e32: f32 = Constants::E;
   /// let x32: Radian<f32> = Radian::from(1.0f32);
   ///
   /// let f_val32 = x32.tanh();
   /// let g_val32 = (1.0f32 - e32.powi(-2i32))/(1.0f32 + e32.powi(-2i32));
   /// let abs_difference32 = (f_val32 - g_val32).abs();
   ///
   /// let e64: f64 = Constants::E;
   /// let x64: Radian<f64> = Radian::from(1.0f64);
   ///
   /// let f_val64 = x64.tanh();
   /// let g_val64 = (1.0f64 - e64.powi(-2i32))/(1.0f64 + e64.powi(-2i32));
   /// let abs_difference64 = (f_val64 - g_val64).abs();
   ///
   /// // Solving tanh() at 1 gives `(1 - e^(-2))/(1 + e^(-2))`
   /// //assert!(abs_difference32 < 1.0e-10);
   /// assert!(abs_difference64 < 1.0e-10);
   /// ```
   pub fn tanh(&self) -> T
   {
      self.value.tanh()
   }
}

impl<T> ::std::ops::Deref for Radian<T> where T: Real
{
   type Target = T;

   fn deref(&self) -> &T
   {
      &self.value
   }
}

impl<T> ::std::ops::DerefMut for Radian<T> where T: Real
{
   fn deref_mut<'a>(&'a mut self) -> &'a mut T
   {
      &mut self.value
   }
}

binary_operator_impl!(Add::add, Radian {value}, Real);
binary_operator_impl!(Sub::sub, Radian {value}, Real);
binary_operator_impl!(Mul::mul, Radian {value}, Real);
binary_operator_impl!(Div::div, Radian {value}, Real);
binary_operator_impl!(Rem::rem, Radian {value}, Real);

impl<T> ::std::ops::Neg for Radian<T>
   where T: Neg<Output = T>
{
   type Output = Radian<T>;

   fn neg(self) -> Radian<T>
   {
      let mut radians: Radian<T>;

      radians = self;
      radians.value = radians.value.neg();
      radians
   }
}

binary_operator_assign_impl!(AddAssign::add_assign,
                             Radian {value}, Real);
binary_operator_assign_impl!(SubAssign::sub_assign,
                             Radian {value}, Real);
binary_operator_assign_impl!(MulAssign::mul_assign,
                             Radian {value}, Real);
binary_operator_assign_impl!(DivAssign::div_assign,
                             Radian {value}, Real);
binary_operator_assign_impl!(RemAssign::rem_assign,
                             Radian {value}, Real);

impl ::std::convert::From<Radian<f32>> for f32
{
   fn from(radians: Radian<f32>) -> f32
   {
      radians.value
   }
}

impl ::std::convert::From<Radian<f64>> for f64
{
   fn from(radians: Radian<f64>) -> f64
   {
      radians.value
   }
}

impl<T> ::std::convert::From<T> for Radian<T> where T: Real
{
   fn from(val: T) -> Radian<T>
   {
      Radian::new(val)
   }
}

impl<T> ::std::convert::From<Degree<T>> for Radian<T>
   where T: Real
{
   fn from(degrees: Degree<T>) -> Radian<T>
   {
      let rads: T;

      rads = degrees.value * Constants::PI_DIVIDED_BY_180;
      Radian::new(rads)
   }
}

impl<T> ::std::fmt::Debug for Radian<T>
   where T: Real
{
   fn fmt(&self, f: &mut ::std::fmt::Formatter) -> ::std::fmt::Result
   {
      ::std::fmt::Display::fmt(self, f)
   }
}

impl<T> ::std::fmt::Display for Radian<T>
   where T: Real
{
   fn fmt(&self, f: &mut ::std::fmt::Formatter) -> ::std::fmt::Result
   {
      write!(f, "{}", self.value)
   }
}
